﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

using Platform.ComponentModel;

namespace Platform.Data
{
    public class DataPagingViewModel<TClient,TModel> : DisposableAndObservableBase
        where TClient : IDataCollection<TModel>
    {
        private CancellationTokenSource _CancellationTokenSource;
        private CancellationToken _CancellationToken;
        private TaskScheduler _TaskScheduler;

        public DataPagingViewModel()
        {
            _TaskScheduler = TaskScheduler.FromCurrentSynchronizationContext();
        }

        public DataPagingViewModel(TClient client)
            : this()
        {
            Client = client;
        }

        public TClient Client { get; set; }

        #region Result Set

        public List<TModel> Data
        {
            get { return _Data; }
            protected set { SetPropertyValue(ref _Data, value, nameof(Data)); }
        }
        private List<TModel> _Data = new List<TModel>();

        public long TotalCount
        {
            get { return _TotalCount; }
            protected set { SetPropertyValue(ref _TotalCount, value, nameof(TotalCount)); }
        }
        private long _TotalCount;

        public int FilteredCount
        {
            get { return _FilteredCount; }
            protected set { SetPropertyValue(ref _FilteredCount, value, nameof(FilteredCount)); }
        }
        private int _FilteredCount;

        public int PageCount
        {
            get { return _PageCount; }
            protected set { SetPropertyValue(ref _PageCount, value, nameof(PageCount)); }
        }
        private int _PageCount;

        #endregion

        #region Filter and Sort

        public string FilterExpression
        {
            get { return _FilterExpression; }
            set { if (SetPropertyValue(ref _FilterExpression, value, nameof(FilterExpression))) LoadPage(); }
        }
        private string _FilterExpression;

        public string SortExpression
        {
            get { return _SortExpression; }
            set { if (SetPropertyValue(ref _SortExpression, value, nameof(SortExpression))) LoadPage(); }
        }
        private string _SortExpression;

        #endregion

        #region Paging Properties

        public int PageIndex
        {
            get { return _PageIndex; }
            set { if (SetPropertyValue(ref _PageIndex, value, nameof(PageIndex))) LoadPage(); }
        }
        private int _PageIndex;

        public int PageSize
        {
            get { return _PageSize; }
            set { if( SetPropertyValue(ref _PageSize, value, nameof(PageSize))) LoadPage(); }
        }
        private int _PageSize = 10;

        #endregion

        #region Loading

        /// <summary>
        /// Page load failed
        /// </summary>
        public event ExceptionEventHandler PageLoadFailed;

        /// <summary>
        /// Signals page is loading
        /// </summary>
        public event EventHandler PageLoading;

        /// <summary>
        /// Signals page was loaded
        /// </summary>
        public event EventHandler PageLoaded;

        public bool IsLoading
        {
            get { return _IsLoading; }
            set { SetPropertyValue(ref _IsLoading, value, nameof(IsLoading)); }
        }
        private bool _IsLoading;


        /// <summary>
        /// Cancel loading page
        /// </summary>
        public void Cancel()
        {
            if (_IsLoading)
                _CancellationTokenSource.Cancel();
        }

        /// <summary>
        /// Load a page
        /// </summary>
        public void LoadPage()
        {
            // check if isloading
            if( _IsLoading )
            {
                _IsReloadRequested = true;
                return;
            }

            // create cancellation token source
            _CancellationTokenSource = new CancellationTokenSource();
            _CancellationToken = _CancellationTokenSource.Token;

            // page reloaded
            PageLoading?.Invoke(this, EventArgs.Empty);

            // start loading
            Client
                .GetAsync(PageIndex * PageSize, PageSize, FilterExpression, SortExpression, _CancellationToken)
                .ContinueWith( OnPageLoaded, _CancellationToken, TaskContinuationOptions.NotOnFaulted, _TaskScheduler );

            // set loading
            _IsLoading = true;
            _IsReloadRequested = false;
        }
        private bool _IsReloadRequested;

        /// <summary>
        /// Handle page loaded
        /// </summary>
        /// <param name="task"></param>
        /// <returns></returns>
        protected virtual void OnPageLoaded( Task<DataCollectionResult<TModel>> task )
        {
            // done loading
            _IsLoading = false;

            // no cancellation option
            _CancellationTokenSource.Dispose();
            _CancellationTokenSource = null;

            // check if reload requested
            if (_IsReloadRequested)
            {
                _IsReloadRequested = false;
                LoadPage();
                return;
            }

            // task canceled
            if (task.IsCanceled)
                return;

            // show failure (make this event driven)
            if (task.IsFaulted)
            {
                PageLoadFailed?.Invoke(this, new ExceptionEventArgs(task.Exception));
                return;
            }

            // set results
            var result = task.Result;
            TotalCount = result.TotalCount;
            FilteredCount = result.FilteredCount;
            PageCount = (int)((result.FilteredCount + PageSize - 1) / PageSize);
            Data = result.Data;

            // page reloaded
            PageLoaded?.Invoke(this, EventArgs.Empty);
        }

        #endregion
    }
}
