﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;

using System.Net;
using System.Net.Sockets;

namespace Platform.Extensions
{
    public static class IPAddressExtensions
    {
        #region From String To IPAddress

        /// <summary>
        /// Parse IPV4
        /// </summary>
        /// <param name="source"></param>
        /// <param name="address"></param>
        /// <returns></returns>
        private static byte[] ToIP4(this string source)
        {
            // check IPv4 format
            string[] items = source.Split('.');
            if (items.Length != 4) return null;

            // get bytes
            Byte? p1 = items[0].ToByte(10, null);
            Byte? p2 = items[1].ToByte(10, null);
            Byte? p3 = items[2].ToByte(10, null);
            Byte? p4 = items[3].ToByte(10, null);

            // check bytes
            if (p1 == null || p2 == null || p3 == null || p4 == null) return null;

            // create address
            return new byte[] { p1.Value, p2.Value, p3.Value, p4.Value };
        }

        /// <summary>
        /// Parse IPV6
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        private static byte[] ToIP6(this string source)
        {
            string[] parts = source.Split(':');
            byte[] result = new byte[16];
            int remain = parts.Length;
            int target = 0;
            int count = 16;
            int part = 0;

            // last part is IPv4, get it
            if (parts[remain - 1].Contains('.'))
            {
                byte[] ip4 = ToIP4(parts[remain - 1]);
                if (ip4 == null) return null;
                Array.Copy(ip4, 0, result, 12, 4);
                count -= 4;
                remain--;
            }

            // parse ipv6 (::ffff:192.168.89.9)
            // 4 parts
            // count = 12
            // remain = 3 (0-2)
            for (; part < remain; part++)
            {
                // check for empty part (0)
                if (parts[part].Length == 0)
                {
                    while (target < count - 2 * (remain - part - 1)) result[target++] = 0;
                    continue;
                }

                // parse 16 bits
                ushort? v = parts[part].ToUInt16(16);
                if (v == null) return null;

                // store part
                result[target++] = (byte)(v.Value >> 8);
                result[target++] = (byte)(v.Value & 0xFF);
            }

            // return result
            return result;
        }

        /// <summary>
        /// Parse endpoint
        /// </summary>
        public static IPAddress ToIPAddress(this string source) { return ToIPAddress(source, null); }

        /// <summary>
        /// Parse endpoint
        /// </summary>
        public static IPAddress ToIPAddress(this string source, IPAddress address)
        {
            // check source
            if (source == null || source.Length == 0)
                return null;

            // check for IPv6
            if (source.Contains(':'))
            {
                byte[] result = ToIP6(source);
                if (result == null) return address;
                return new IPAddress(result);
            }

            // check for IPv4
            if (source.Contains('.'))
            {
                byte[] result = ToIP4(source);
                if (result == null) return address;
                return new IPAddress(result);
            }

            // return default address
            return address;
        }

        #endregion

        /// <summary>
        /// Returns true if address is multicast
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static bool IsMulticast(this IPAddress address)
        {
            if (address == null) return false;

            if (address.AddressFamily == AddressFamily.InterNetworkV6)
                return address.GetAddressBytes()[0] == 0xFF;
            if (address.AddressFamily == AddressFamily.InterNetwork)
                return address.GetAddressBytes()[0] >= 224 && address.GetAddressBytes()[0] <= 239;
            return false;
        }

        /// <summary>
        /// Returns true if address is valid for binding
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static bool IsUnicast(this IPAddress address)
        {
            if (address == null) return false;
            if (address.IsMulticast()) return false;

            if (address.AddressFamily == AddressFamily.InterNetworkV6)
            {
                if (address.Equals(IPAddress.IPv6Any)) return false;
                if (address.Equals(IPAddress.IPv6None)) return false;
                return true;
            }

            if (address.AddressFamily == AddressFamily.InterNetwork)
            {
                if (address.Equals(IPAddress.Any)) return false;
                if (address.Equals(IPAddress.Broadcast)) return false;
                if (address.Equals(IPAddress.None)) return false;
                return true;
            }

            return false;
        }

        /// <summary>
        /// Returns true if address is valid for binding
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public static bool IsValidForBinding(this IPAddress address)
        {
            if (address == null)
                return false;

            if (address.AddressFamily == AddressFamily.InterNetworkV6)
            {
                if (address.Equals(IPAddress.IPv6Any)) return false;
                if (address.Equals(IPAddress.IPv6Loopback)) return false;
                if (address.Equals(IPAddress.IPv6None)) return false;
                return true;
            }

            if (address.AddressFamily == AddressFamily.InterNetwork)
            {
                if (address.Equals(IPAddress.Any)) return false;
                if (address.Equals(IPAddress.Broadcast)) return false;
                if (address.Equals(IPAddress.Loopback)) return false;
                if (address.Equals(IPAddress.None)) return false;
                return true;
            }

            return false;
        }
    }
}
