﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.NetworkInformation;

namespace Platform.Extensions
{
    public static class PhysicalAddressExtensions
    {
        #region Special Addresses

        private static PhysicalAddress _Broadcast = new PhysicalAddress(new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF });
        private static PhysicalAddress _None = new PhysicalAddress(new byte[] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 });

        /// <summary>
        /// Get the broadcast physical address (FF:FF:FF:FF:FF:FF)
        /// </summary>
        public static PhysicalAddress Broadcast
        {
            get { return _Broadcast; }
        }

        /// <summary>
        /// Get the empty physical address (00:00:00:00:00:00)
        /// </summary>
        public static PhysicalAddress None
        {
            get { return _None; }
        }

        #endregion

        #region Format PhysicalAddress

        /// <summary>
        /// Convert Physical Address to string
        /// </summary>
        /// <param name="address"></param>
        /// <param name="format"></param>
        /// <param name="provider"></param>
        /// <returns></returns>
        public static string ToString(this PhysicalAddress address, string format, IFormatProvider provider)
        {
            // no address
            if (address == null)
                return null;

            // default formatting
            if (string.IsNullOrEmpty(format))
                return address.ToString();

            // use custom provider
            if (provider != null)
            {
                ICustomFormatter formatter = provider.GetFormat(typeof(ICustomFormatter)) as ICustomFormatter;
                return formatter.Format(format, address, null);
            }

            // use default provider
            if (format.ToUpper().StartsWith("X"))
            {
                // create result
                StringBuilder result = new StringBuilder();
                foreach (byte value in address.GetAddressBytes())
                {
                    // insert seperator
                    if (result.Length > 0 && format.Length>1) 
                        result.Append(format[1]);

                    // insert characters
                    string item = string.Format("{0:X2}", value);
                    result.AppendFormat(format[0] == 'X' ? item.ToUpper() : item.ToLower());
                }

                // as string
                return result.ToString();
            }

            // use default provider
            if (format.ToUpper().StartsWith("X"))
            {
                // default use ':' seperator
                if (format.Length == 1) format += ":";

                // create result
                StringBuilder result = new StringBuilder();
                foreach (byte value in address.GetAddressBytes())
                {
                    if (result.Length > 0) result.Append(format[1]);
                    string item = string.Format("{0:X2}", value);
                    result.AppendFormat(format[0] == 'X' ? item.ToUpper() : item.ToLower());
                }

                // as string
                return result.ToString();
            }

            // invalid arguments
            throw new ArgumentException("Invalid formatting string");
        }

        #endregion

        #region To PhysicalAddress

        /// <summary>
        /// Parse PhysicalAddress
        /// </summary>
        public static PhysicalAddress ToPhysicalAddress(this string source) { return ToPhysicalAddress(source, null); }

        /// <summary>
        /// Parse PhysicalAddress
        /// </summary>
        public static PhysicalAddress ToPhysicalAddress(this string source, PhysicalAddress address)
        {
            // check source
            if (source == null || source.Length == 0)
                return address;

            // get compact source
            var text = new string( source.ToUpper().ToCharArray().Where(c => char.IsDigit(c) || (c >= 'A' && c <= 'F')).ToArray() );
            if (text.Length != 12) return address;

            // get physical address
            return new PhysicalAddress(
                new byte[]
                {
                    text.Substring(0, 2).ToByte(16).Value,
                    text.Substring(2, 2).ToByte(16).Value,
                    text.Substring(4, 2).ToByte(16).Value,
                    text.Substring(6, 2).ToByte(16).Value,
                    text.Substring(8, 2).ToByte(16).Value,
                    text.Substring(10, 2).ToByte(16).Value
                });            
        }

        #endregion
    }
}
