﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Platform.Extensions
{
    public static class StringExtensions
    {
        #region URI Encode / Decode

        /// <summary>
        /// Decode Uri string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="plainset"></param>
        /// <returns></returns>        
        public static string UriDecode(this string source, bool isformencoded)
        {
            // no decode on null value
            if (source == null) return null;

            // get result builder
            StringBuilder result = new StringBuilder();

            // process source
            for (int i = 0; i < source.Length; i++)
            {
                // plus sign
                if (isformencoded && source[i] == '+')
                {
                    result.Append(' ');
                    continue;
                }

                // decode character
                if (source[i] != '%')
                {
                    result.Append(source[i]);
                    continue;
                }

                // check format
                if (i + 2 >= source.Length)
                    throw new Exception("Invalid Uri Format");

                // check for failue
                var code = source.Substring(i + 1, 2).ToInt32(16, -1);
                if (code == -1) throw new Exception("Invalid Uri Format");

                result.Append((char)code);
                i += 2;
            }

            // return result
            return result.ToString();
        }

        /// <summary>
        /// Encode Uri string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="plainset"></param>
        /// <returns></returns>
        public static string UriEncode(this string source, string plainset, bool useformencoding)
        {
            // no encode on null value
            if (source == null) return null;

            // get result builder
            StringBuilder result = new StringBuilder();

            // process source
            for (int i = 0; i < source.Length; i++)
            {
                // plus sign
                if (useformencoding && source[i] == ' ')
                {
                    result.Append('+');
                    continue;
                }

                // escaping
                int offset = plainset.IndexOf(source[i]);
                if (offset != -1) result.Append(source[i]);
                else result.AppendFormat("%{0:X2}", (byte)source[i]);
            }

            // return result
            return result.ToString();
        }

        #endregion

        #region JSON Encode / Decode

        /// <summary>
        /// Check if escape needed
        /// </summary>
        /// <param name="src"></param>
        /// <param name="i"></param>
        /// <returns></returns>
        private static bool NeedEscape(this string src, int i)
        {
            char c = src[i];
            return c < 32 || c == '"' || c == '\\'
                // Broken lead surrogate
                || (c >= '\uD800' && c <= '\uDBFF' &&
                    (i == src.Length - 1 || src[i + 1] < '\uDC00' || src[i + 1] > '\uDFFF'))
                // Broken tail surrogate
                || (c >= '\uDC00' && c <= '\uDFFF' &&
                    (i == 0 || src[i - 1] < '\uD800' || src[i - 1] > '\uDBFF'))
                // To produce valid JavaScript
                || c == '\u2028' || c == '\u2029'
                // Escape "</" for <script> tags
                || (c == '/' && i > 0 && src[i - 1] == '<');
        }

        /// <summary>
        /// Convert to JSON string
        /// </summary>
        /// <param name="src"></param>
        /// <returns></returns>
        public static string EncodeJsonString(this string src)
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder();

            int start = 0;
            for (int i = 0; i < src.Length; i++)
                if (NeedEscape(src, i))
                {
                    sb.Append(src, start, i - start);
                    switch (src[i])
                    {
                        case '\b': sb.Append("\\b"); break;
                        case '\f': sb.Append("\\f"); break;
                        case '\n': sb.Append("\\n"); break;
                        case '\r': sb.Append("\\r"); break;
                        case '\t': sb.Append("\\t"); break;
                        case '\"': sb.Append("\\\""); break;
                        case '\\': sb.Append("\\\\"); break;
                        case '/': sb.Append("\\/"); break;
                        default:
                            sb.Append("\\u");
                            sb.Append(((int)src[i]).ToString("x04"));
                            break;
                    }
                    start = i + 1;
                }
            sb.Append(src, start, src.Length - start);
            return sb.ToString();
        }

        #endregion

        /// <summary>
        /// Make valid filename
        /// </summary>
        /// <param name="source"></param>
        /// <param name="replacementChar"></param>
        /// <returns></returns>
        public static string MakeValidFileName(this string source, char replacementChar = '_')
        {
            var invalidChars = new HashSet<char>(Path.GetInvalidFileNameChars());
            return new string(source.Select(c => invalidChars.Contains(c) ? replacementChar : c).ToArray());
        }

        /// <summary>
        /// Make valid filename
        /// </summary>
        /// <param name="source"></param>
        /// <param name="replacementChar"></param>
        /// <returns></returns>
        public static string MakeValidFilePath(this string source, char replacementChar = '_')
        {
            var invalidChars = new HashSet<char>(Path.GetInvalidPathChars());
            return new string(source.Select(c => invalidChars.Contains(c) ? replacementChar : c).ToArray());
        }

        public static bool Contains(this string text, string search, StringComparison comparison)
        {
            if (text == null || search == null) return false;
            if (string.IsNullOrEmpty(text)) return false;
            if (string.IsNullOrEmpty(search)) return false;
            return text.IndexOf(search, comparison) >= 0;
        }

        /// <summary>
        /// Performs a case insensitive compare
        /// </summary>
        /// <param name="source"></param>
        /// <param name="target"></param>
        /// <returns></returns>
        public static bool IsEqualTo(this string source, string target)
        {
            return IsEqualTo(source, target, StringComparison.CurrentCultureIgnoreCase);
        }

        /// <summary>
        /// Check if two strings are equal
        /// </summary>
        /// <param name="source"></param>
        /// <param name="target"></param>
        /// <returns></returns>
        public static bool IsEqualTo(this string source, string target, bool ignoreCase)
        {
            if (ignoreCase) return IsEqualTo(source, target, StringComparison.CurrentCultureIgnoreCase);
            else return IsEqualTo(source, target, StringComparison.CurrentCulture);
        }

        /// <summary>
        /// Check if two strings are equal
        /// </summary>
        /// <param name="source"></param>
        /// <param name="target"></param>
        /// <returns></returns>
        public static bool IsEqualTo(this string source, string target, StringComparison comparison )
        {
            if (source == null && target == null) return true;
            if (source == null && target != null) return false;
            if (source != null && target == null) return false;
            return String.Compare(source, target, comparison) == 0;
        }

        /// <summary>
        /// Check if two strings are equal
        /// </summary>
        /// <param name="source"></param>
        /// <param name="target"></param>
        /// <returns></returns>
        public static bool IsEqualToAfterTrim(this string source, string target)
        {
            return string.Compare(source?.Trim(), target?.Trim(), StringComparison.CurrentCultureIgnoreCase) == 0;
        }

        /// <summary>
        /// Check if string is contained in string collection collection
        /// </summary>
        public static bool IsContainedAfterTrim(this string source, IEnumerable<string> target)
        {
            if (target == null) return false;
            return target.Any(t => t.IsEqualToAfterTrim(source));
        }

        /// <summary>
        /// Compares two string if they have the same starting characters
        /// </summary>
        /// <param name="source"></param>
        /// <param name="target"></param>
        /// <param name="ignorecase"></param>
        /// <returns></returns>
        public static bool StartEquals(this string source, string target, bool ignorecase)
        {
            int length = Math.Min(source.Length, target.Length);
            if (ignorecase)
            {
                for (int i = 0; i < length; i++)
                    if (char.ToUpper(source[i]) != char.ToUpper(target[i]))
                        return false;
            }
            else
            {
                for (int i = 0; i < length; i++)
                    if (source[i] != target[i])
                        return false;
            }
            return true;
        }

        /// <summary>
        /// Converts a string 
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public static IEnumerable<string> SplitLines(this string message, bool ignoreEmptyLines)
        {
            // no message
            if (message == null)
                yield break;

            // process message
            using (StringReader reader = new StringReader(message))
                while (true)
                {
                    // read line
                    var text = reader.ReadLine();
                    if (text == null) break;

                    // get actual text
                    text = text.Replace("\r", "").Replace("\n", "");
                    if (StringExtensions.IsNullOrWhiteSpace(text) && ignoreEmptyLines) continue;

                    // return result
                    yield return text;
                }
        }

        /// <summary>
        /// Block Format Code Text
        /// </summary>
        /// <param name="source"></param>
        /// <param name="prefix"></param>
        /// <returns></returns>
        public static string BlockFormat(this string source, string prefix)
        {
            // normalize line endings
            source = source.Replace("\r\n", "\n").Replace("\r", "\n");

            var result = new StringBuilder();
            foreach (var line in source.Split('\n'))
                result.AppendLine(line.Trim('\r', '\n', '\t', ' ').Insert(0, prefix));
            return result.ToString();
        }

        /// <summary>
        /// Format list of strings as a bullit list 
        /// </summary>
        /// <param name="format"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        public static IEnumerable<string> FormatList(this IEnumerable<string> source, string prefix, params object[] args)
        {
            // create formatted prefix
            prefix = string.Format(prefix, args);
            string margin = new string(' ', prefix.Length);
            var header = source.Take(1).Select(s => prefix + s);
            var message = source.Skip(1).Select(s => margin + s);
            return header.Concat(message);            
        }

        /// <summary>
        /// Convert collection of strings to a single string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="seperator"></param>
        /// <returns></returns>
        public static string ToString(this IEnumerable<string> source, string seperator)
        {
            var text = source.Where( s => !string.IsNullOrEmpty(s) ).ToArray();
            return string.Join(seperator, text);
        }

        /// <summary>
        /// Split string into key and value
        /// </summary>
        /// <param name="source"></param>
        /// <param name="seperator"></param>
        /// <param name="key"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static KeyValuePair<string, string> SplitKeyAndValue(this string source, char seperator )
        {
            int index = source.IndexOf(seperator);
            if (index == -1) return new KeyValuePair<string, string>(source,"");
            else return new KeyValuePair<string,string>( source.Substring(0, index).Trim(), source.Substring(index + 1) );
        }

        /// <summary>
        /// Check if string matches pattern
        /// </summary>
        /// <param name="source"></param>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public static bool PatternMatch(this string source, string pattern, bool ignorecase)
        {
            // check arguments
            if (source == null || pattern == null) return false;

            // check for ignore case
            if( ignorecase )
            {
                source = source.ToUpper();
                pattern = pattern.ToUpper();
            }

            // initialize
            int pindex = 0;
            int sindex = 0;

            // math pattern
            for (; sindex < source.Length && pindex < pattern.Length; sindex++)
            {
                // check pattern
                switch (pattern[pindex])
                {
                    // wildcard compare
                    case '*':
                        if ((pindex + 1 < pattern.Length && source[sindex] == pattern[pindex + 1]) || 
                            sindex == source.Length - 1)
                            pindex++;
                        break;

                    // wildcard compare
                    case '?':
                        pindex++;
                        break;

                    // normal compare
                    default:
                        if (source[sindex] != pattern[pindex]) return false;
                        pindex++;
                        break;
                }
            }

            // success
            return (sindex == source.Length && pindex == pattern.Length);
        }

        /// <summary>
        /// Split string into a key value pair
        /// </summary>
        /// <param name="source"></param>
        /// <param name="seperators"></param>
        /// <returns></returns>
        public static string[] SplitOnce(this string source, params char[] seperators)
        {
            int index = source.IndexOfAny(seperators);
            if (index == -1) return new string[] { source, "" };
            return new string[] { source.Substring(0, index), source.Substring(index + 1) };
        }

        /// <summary>
        /// Find position of first character not in the given characterset,
        /// or Length if all characters are in the characterset
        /// </summary>
        public static int IndexOfOther(this string source, string chars)
        {
            return IndexOfOther(source, chars.ToCharArray());
        }

        /// <summary>
        /// Find position of first character not in the given characterset,
        /// or Length if all characters are in the characterset
        /// </summary>
        public static int IndexOfOther(this string source, char[] chars)
        {
            return IndexOfOther(source, chars, 0);
        }

        /// <summary>
        /// Find position of first character not in the given characterset,
        /// or Length if all characters are in the characterset
        /// </summary>
        public static int IndexOfOther(this string source, string chars, int startIndex)
        {
            return IndexOfOther(source, chars.ToCharArray(), startIndex);
        }

        /// <summary>
        /// Remove all character in the specified set from the string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="chars"></param>
        /// <returns></returns>
        public static string Filter(this string source, params char[] chars)
        {
            StringBuilder result = new StringBuilder();
            foreach (var item in source)
                if (Array.IndexOf(chars, item) == -1)
                    result.Append(item);
            return result.ToString();
        }

        /// <summary>
        /// Find position of first character not in the given characterset,
        /// or Length if all characters are in the characterset
        /// </summary>
        public static int IndexOfOther(this string source, char[] chars, int startIndex)
        {
            for (int index = startIndex; index < source.Length; index++)
                if (!chars.Contains(source[index]))
                    return index;
            return source.Length;
        }

        /// <summary>
        /// Take a part from the source
        /// </summary>
        /// <param name="source"></param>
        /// <param name="seperator"></param>
        /// <returns></returns>
        public static string TakePart( this StringBuilder source, string seperator )
        {
            // get seperator seach
            var result = source.ToString();
            int index = result.IndexOf(seperator);

            // return last part
            if (index == -1)
            {
                source.Length = 0;
                return result;
            }

            // return part
            source.Remove(0, index + seperator.Length);
            return result.Substring(0, index);
        }

        /// <summary>
        /// Checks for null, empty or white space string
        /// </summary>
        /// <param name="text"></param>
        /// <returns></returns>
        public static bool IsNullOrWhiteSpace(this string text)
        {
            if (string.IsNullOrEmpty(text)) return true;
            foreach (var c in text) if (!char.IsWhiteSpace(c)) return false;
            return true;
        }

        /// <summary>
        /// Search string builder for text
        /// </summary>
        /// <param name="source"></param>
        /// <param name="search"></param>
        /// <returns></returns>
        public static int SearchAfter(this StringBuilder source, string search)
        {
            return SearchAfter(source, 0, search);
        }

        /// <summary>
        /// Gets text until contents
        /// </summary>
        /// <param name="source"></param>
        /// <param name="contents"></param>
        /// <param name="caseinsensitive"></param>
        /// <returns></returns>
        public static int SearchAfter(this StringBuilder source, int offset, string search)
        {
            // need some source
            if (search.Length == 0) return -1;
            if (source.Length - offset < search.Length) return -1;

            // search into buffer
            for (int i = offset; i <= source.Length - search.Length; i++ )
            {
                // search contents
                int j = 0;
                for (; j < search.Length && source[i + j] == search[j]; j++) ;

                // found
                if (j == search.Length) return i + search.Length;
            }

            // not found
            return -1;
        }

        /// <summary>
        /// Search string builder for text
        /// </summary>
        /// <param name="source"></param>
        /// <param name="search"></param>
        /// <returns></returns>
        public static int IndexOfEnd(this StringBuilder source, string search)
        {
            return IndexOfEnd(source, search, 0);
        }

        /// <summary>
        /// Gets text until contents
        /// </summary>
        /// <param name="source"></param>
        /// <param name="contents"></param>
        /// <param name="caseinsensitive"></param>
        /// <returns></returns>
        public static int IndexOfEnd(this StringBuilder source, string search, int offset)
        {
            int result = IndexOf(source, search, offset);
            if (result != -1) result += search.Length;
            return result;
        }

        public static int IndexOf(this StringBuilder source, string search)
        {
            return IndexOf(source, search, 0);
        }

        /// <summary>
        /// Gets text until contents
        /// </summary>
        /// <param name="source"></param>
        /// <param name="contents"></param>
        /// <param name="caseinsensitive"></param>
        /// <returns></returns>
        public static int IndexOf(this StringBuilder source, string search, int offset)
        {
            // need some source
            if (search.Length == 0) return -1;
            if (source.Length - offset < search.Length) return -1;

            // search into buffer
            for (int i = offset; i <= source.Length - search.Length; i++)
            {
                // search contents
                int j = 0;
                for (; j < search.Length && source[i + j] == search[j]; j++) ;

                // found
                if (j == search.Length) return i;
            }

            // not found
            return -1;
        }

        /// <summary>
        /// Truncate string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="length"></param>
        /// <param name="trim"></param>
        /// <returns></returns>
        public static string Truncate(this string source, int length, string trim = "...")
        {
            if (source == null) return null;
            if (source.Length <= length) return source;
            return source.Substring(0, length - trim.Length) + trim;
        }
    }
}
