﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;

namespace Platform.Net
{
    /// <summary>
    /// A writer that can write miscellanious
    /// values, mostly network protocol related,
    /// to a text string
    /// </summary>
    public class NetStringWriter
    {
        /// <summary>
        /// The actual result
        /// </summary>
        private StringBuilder _Target;

        /// <summary>
        /// Construct a new writer
        /// </summary>
        public NetStringWriter()
        {
            _Target = new StringBuilder();
        }

        /// <summary>
        /// Writes a string to the result
        /// </summary>
        /// <param name="data"></param>
        public void Write(string data)
        {
            _Target.Append(data);
        }

        /// <summary>
        /// Writes a character to the result
        /// </summary>
        /// <param name="data"></param>
        public void Write(char data)
        {
            _Target.Append(data);
        }

        /// <summary>
        /// Writes a formatted string to the result
        /// </summary>
        /// <param name="format"></param>
        /// <param name="args"></param>
        public void WriteFormat(string format, params object[] args)
        {
            _Target.AppendFormat(format, args);
        }

        /// <summary>
        /// Writers a token or escaped quoted string to the result
        /// </summary>
        /// <param name="value"></param>
        public void WriteTokenOrQuotedString( string value )
        {
            if (value == null) return;
            if (NetValue.IsToken(value)) _Target.Append(value);
            else _Target.AppendFormat("\"{0}\"", value);
        }

        /// <summary>
        /// Convert to RFC 822 string
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string WriteDateTimeHttp(DateTime value)
        {
            return value.ToUniversalTime().ToString("ddd, d MMM yyyy HH:mm:ss GMT", CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Convert to RFC 822 string
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string WriteDateTimeRfc822(DateTime value)
        {
            return value.ToUniversalTime().ToString("ddd, d MMM yy HH:mm:ss UT", CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Convert to RFC 1123 string
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string WriteDateTimeRfc1123(DateTime value)
        {
            return value.ToUniversalTime().ToString("ddd, d MMM yyyy HH:mm:ss UT", CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Convert to RFC 850 string
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string WriteDateTimeRfc850(DateTime value)
        {
            return value.ToUniversalTime().ToString("dddd, dd-MMM-yy HH:mm:ss UTC", CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Convert to asctime string
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string WriteDateTimeAscTime(DateTime value)
        {
            return value.ToUniversalTime().ToString("ddd MMM dd HH:mm:ss yyyy", CultureInfo.InvariantCulture);
        }

        /// <summary>
        /// Gets the result as a string
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return _Target.ToString();
        }
    }
}
