﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace Platform.Schedules
{
    public static class TimeRosterExtensions
    {
        public const int MinutesPerDay = 60 * 24;
        public const int MinutesPerWeek = MinutesPerDay * 7;

        /// <summary>
        /// Returns true if period is active at the given date
        /// </summary>
        /// <param name="roster"></param>
        /// <param name="date"></param>
        /// <returns></returns>
        public static bool IsActiveAt( this ITimeWeekRoster roster, DateTime date )
        {
            // get start of the week
            var weekstart = date.Subtract(TimeSpan.FromMinutes(((int)date.DayOfWeek) * MinutesPerDay));
            var weekoffset = date - weekstart;

            // not active within date schedule
            if (roster.StartDate > date || roster.UntilDate < date)
                return false;

            // not active within week schedule
            if (roster.StartMinute > weekoffset || roster.UntilMinute < weekoffset)
                return false;

            // return valid
            return true;
        }

        /// <summary>
        /// Gets the duration from the current date upto the next change to
        /// the requested state
        /// </summary>
        /// <param name="roster"></param>
        /// <param name="date"></param>
        /// <param name="state"></param>
        /// <returns></returns>
        public static TimeSpan? GetDurationUntilNextChange(this ITimeWeekRoster roster, DateTime date, bool? state)
        {
            // get start of the week
            var weekstart = date.Subtract(TimeSpan.FromMinutes(((int)date.DayOfWeek) * MinutesPerDay));

            DateTime startdate = weekstart + roster.StartMinute;
            if (startdate < date) startdate += TimeSpan.FromMinutes(MinutesPerWeek);

            DateTime untildate = weekstart + roster.UntilMinute;
            if (untildate < date) untildate += TimeSpan.FromMinutes(MinutesPerWeek);

            // check if week dates are out of bounds with period dates 
            // WS (weekstart), WE (weekend), DS (date start), DE (date end)
            // before:  WS-WE  DS-DS  
            // after:   DS-DE  WS-WE
            if (startdate > roster.UntilDate) return null;
            if (untildate < roster.StartDate) return null;

            // check if dates overlap 
            // WS DS WE DE
            // DS WS DE WE
            // WS DS DE WE
            if (startdate < roster.StartDate) startdate = roster.StartDate;
            if (untildate > roster.UntilDate) untildate = roster.UntilDate;

            // get first start of this period
            if (state == true) return startdate - date;
            if (state == false) return untildate - date;
            return Min(startdate, untildate) - date;
        }

        /// <summary>
        /// Gets the next available
        /// </summary>
        /// <returns></returns>
        public static IEnumerable<TimeSpan> GetDurationsUntilNextChange(this IEnumerable<ITimeWeekRoster> periods, DateTime date, bool? state)
        {
            return periods.Select(s => s.GetDurationUntilNextChange(date, state)).Where(s => s.HasValue).Select(s => s.Value);
        }

        /// <summary>
        /// Gets the next available
        /// </summary>
        /// <returns></returns>
        public static TimeSpan GetDurationUntilNextChange(this IEnumerable<ITimeWeekRoster> periods, DateTime date, bool? state)
        {
            var items = periods.GetDurationsUntilNextChange(date, state).OrderBy(s => s).Take(1).ToList();
            if (items.Count == 0) return Timeout.InfiniteTimeSpan;
            else return items[0];
        }

        /// <summary>
        /// Load Time Rosters from Periods
        /// </summary>
        public static IEnumerable<ITimeWeekRoster> ToRosters( this IEnumerable<ITimeSchedule> periods )
        {
            // get current date
            var now = DateTime.Now;

            // normalize timezones
            foreach (var period in periods)
            {
                // out of range
                if (period.UntilDate != null && period.UntilDate.Value < now)
                    continue;

                // unfold all days
                for (int i = 0; i < 7; i++)
                {
                    // check if day enabled
                    if (!period.IsDayEnabled((DayOfWeek)i))
                        continue;

                    // var roster
                    var roster = new TimeWeekRoster();
                    roster.StartDate = period.StartDate ?? DateTime.MinValue;
                    roster.UntilDate = period.UntilDate ?? DateTime.MaxValue;
                    roster.StartMinute = period.StartTime ?? TimeSpan.Zero;
                    roster.UntilMinute = period.UntilTime ?? TimeSpan.FromMinutes(MinutesPerDay);

                    // forward timing
                    if (roster.UntilMinute < roster.StartMinute)
                        roster.UntilMinute += TimeSpan.FromMinutes(MinutesPerDay);

                    // adjust for day period
                    roster.StartMinute += TimeSpan.FromMinutes(i * MinutesPerDay);
                    roster.UntilMinute += TimeSpan.FromMinutes(i * MinutesPerDay);

                    // next day
                    if (roster.UntilMinute.TotalMinutes <= MinutesPerWeek)
                    {
                        yield return roster;
                        continue;
                    }

                    // var roster
                    var roster2 = new TimeWeekRoster();
                    roster2.StartDate = period.StartDate ?? DateTime.MinValue;
                    roster2.UntilDate = period.UntilDate ?? DateTime.MaxValue;
                    roster2.StartMinute = TimeSpan.Zero;
                    roster2.UntilMinute = roster.UntilMinute - TimeSpan.FromMinutes(MinutesPerWeek);

                    roster.UntilMinute = TimeSpan.FromMinutes(MinutesPerDay);
                    yield return roster;
                    yield return roster2;
                }
            }
        }

        /// <summary>
        /// Merge adjecent time rosters
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="U"></typeparam>
        /// <param name="rosters"></param>
        /// <returns></returns>
        public static IEnumerable<ITimeWeekRoster> MergeAdjecents(this IEnumerable<ITimeWeekRoster> rosters)
        {
            // convert to final list
            var source = rosters.ToList();

            // process source
            for (int i = 0; i < source.Count; i++)
            {
                // get original
                var roster = source[i];

                // check for matchings
                for (int j = i + 1; j < source.Count; j++)
                {
                    // called after remova;
                    sameagain:

                    // dates should match
                    if (source[j].StartDate != source[i].StartDate) continue;
                    if (source[j].UntilDate != source[i].UntilDate) continue;

                    // check for overlap
                    if (source[j].StartMinute > source[i].UntilMinute ||
                        source[j].UntilMinute < source[i].StartMinute) continue;

                    // merge now
                    if (source[j].StartMinute < source[i].StartMinute)
                        source[i].StartMinute = source[j].StartMinute;
                    if (source[j].UntilMinute > source[i].UntilMinute)
                        source[i].UntilMinute = source[j].UntilMinute;

                    // remove item (its merged)
                    source.RemoveAt(j);

                    // same index once mere
                    goto sameagain;
                }

                // return result
                yield return roster;
            }
        }

        /// <summary>
        /// Merge adjecent time rosters
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="U"></typeparam>
        /// <param name="rosters"></param>
        /// <returns></returns>
        public static IEnumerable<ITimePeriod> MergeAdjecents(this IEnumerable<ITimePeriod> rosters)
        {
            // convert to final list
            var source = rosters.ToList();

            // process source
            for (int i = 0; i < source.Count; i++)
            {
                // get original
                var roster = source[i];

                // check for matchings
                for (int j = i + 1; j < source.Count; j++)
                {
                    // do same again
                    sameagain:

                    // check for overlap
                    if (source[j].StartDate > source[i].UntilDate ||
                        source[j].UntilDate < source[i].StartDate) continue;

                    // merge now
                    if (source[j].StartDate < source[i].StartDate)
                        source[i].StartDate = source[j].StartDate;
                    if (source[j].UntilDate > source[i].UntilDate)
                        source[i].UntilDate = source[j].UntilDate;

                    // remove item (its merged)
                    source.RemoveAt(j);

                    // sane again after removal
                    goto sameagain;
                }

                // return result
                yield return roster;
            }
        }

        /// <summary>
        /// Gets the minimum date
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public static DateTime Min( params DateTime[] dates)
        {
            return dates.Min(d => d.ToUniversalTime());
        }

        /// <summary>
        /// Gets the maximum date
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public static DateTime Max( params DateTime[] dates)
        {
            return dates.Max(d => d.ToUniversalTime());
        }

        /// <summary>
        /// Create unfolded schedule
        /// </summary>
        /// <param name="rosters"></param>
        /// <param name="start"></param>
        /// <param name="duration"></param>
        /// <returns></returns>
        public static IEnumerable<ITimePeriod> CreateSchedules( this IEnumerable<ITimeWeekRoster> rosters, DateTime start, TimeSpan duration )
        {
            // get ending date
            var until = start + duration;

            var weekstart = start.Date.Subtract(TimeSpan.FromMinutes(((int)start.Date.DayOfWeek) * MinutesPerDay));
            var weekuntil = until.Date.Add(TimeSpan.FromMinutes(((int)start.Date.DayOfWeek) * MinutesPerDay));

            // unfold all 
            foreach ( var roster in rosters )
            {
                // out of range
                if (roster.StartDate >= until) continue;
                if (roster.UntilDate < start) continue;

                // roster start
                var rosterstart = Max(roster.StartDate, start);
                var rosteruntil = Min(roster.UntilDate, until);
                
                // repeat for all contained weeks
                for( var offset = weekstart; offset < weekuntil; offset += TimeSpan.FromMinutes( MinutesPerWeek ) )
                {
                    // create schedule
                    var schedule = new TimePeriod
                    {
                        StartDate = Max(offset + roster.StartMinute, roster.StartDate, start),
                        UntilDate = Min(offset + roster.UntilMinute, roster.UntilDate, until)
                    };

                    // return schedule
                    yield return schedule;
                }
                    
            }
        }
    }
}
